/*
	File:		CompositeClassDriver.c

	Contains:	Core functionality to Composite Class Driver

	Version:	xxx put version here xxx

	Copyright:	 1997-1999 by Apple Computer, Inc., all rights reserved.

*/

#include <Types.h>
#include <Devices.h>
#include <processes.h>
#include <DriverServices.h>
#include <USB.h>

#include "CompositeClassDriver.h"

usbCompositePBStruct newInterfacesPB;

void InitParamBlock(USBDeviceRef theDeviceRef, USBPB * paramblock)
{
	paramblock->usbReference = theDeviceRef;
	paramblock->usbCompletion = (USBCompletion)CompositeDeviceCompletionProc;
	paramblock->pbLength = sizeof(usbCompositePBStruct);
	paramblock->pbVersion = kUSBCurrentPBVersion;
	paramblock->usb.cntl.WIndex = 0; 			
	paramblock->usbBuffer = nil;		
	paramblock->usbStatus = kUSBNoErr;
	paramblock->usbReqCount = 0;
	paramblock->usbActCount = 0;
	paramblock->usb.cntl.WValue = 0;
	paramblock->usbFlags = 0;
}


Boolean immediateError(OSStatus err)
{
	return((err != kUSBPending) && (err != kUSBNoErr) );
}

void CompositeDeviceInitiateTransaction(USBPB *pb)
{
register usbCompositePBStruct *pCompositePB;
OSStatus myErr;

	pCompositePB = (usbCompositePBStruct *)(pb);

	if (pCompositePB->expertTerminatePending)				// TCC <USB16>
		return;
		
	pCompositePB->transDepth++;
	if ((pCompositePB->transDepth < 0) || (pCompositePB->transDepth > 1))
	{
	
		USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": Illegal Transaction Depth", pCompositePB->pb.usbRefcon);
	}
	
	if (pCompositePB->driverRemovalPending)
	{
		pCompositePB->pb.usbRefcon = kReturnFromDriver;
		return;
	}

	switch(pCompositePB->pb.usbRefcon & ~kRetryTransaction)
	{
		case kResetDevice:
			InitParamBlock(pCompositePB->deviceRef, &pCompositePB->pb);
			pCompositePB->pb.usbRefcon |= kCompletionPending;
			pCompositePB->pb.usbCompletion = (USBCompletion)CompositeDeviceCompletionProc;
			myErr = USBResetDevice(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": kResetDevice - immediate error", myErr);
			}
			break;

		case kGetFullConfiguration0:
			InitParamBlock(pCompositePB->deviceRef, &pCompositePB->pb);
			pCompositePB->pb.usb.cntl.WIndex = 0; 			
			pCompositePB->pb.usbRefcon |= kCompletionPending;
			
			pCompositePB->pb.usbCompletion = (USBCompletion)CompositeDeviceCompletionProc;
			myErr = USBGetFullConfigurationDescriptor(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": USBGetFullConfiguration (#1) - immediate error", myErr);
			}
			break;
		
		case kSetConfig:
			InitParamBlock(pCompositePB->deviceRef, &pCompositePB->pb);
			pCompositePB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBStandard, kUSBDevice);
			
			pCompositePB->pb.usb.cntl.BRequest = kUSBRqSetConfig;
			pCompositePB->pb.usb.cntl.WValue = pCompositePB->pFullConfigDescriptor->configValue; 		/* Use configuration ID value from descriptor */
			
			pCompositePB->pb.usbRefcon |= kCompletionPending;
			
			pCompositePB->pb.usbCompletion = (USBCompletion)CompositeDeviceCompletionProc;
			myErr = USBDeviceRequest(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": kSetConfig - immediate error", myErr);
			}
			break;
			
		case kSetRemoteWakeup:			
			InitParamBlock(pCompositePB->deviceRef, &pCompositePB->pb);
			
			pCompositePB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBOut, kUSBStandard, kUSBDevice);			
			
			pCompositePB->pb.usb.cntl.BRequest = kUSBRqSetFeature;
			pCompositePB->pb.usb.cntl.WValue = kUSBFeatureDeviceRemoteWakeup; 
			pCompositePB->pb.usb.cntl.WIndex = 0;
			
			pCompositePB->pb.usbRefcon |= kCompletionPending;
			
			pCompositePB->pb.usbCompletion = (USBCompletion)CompositeDeviceCompletionProc;
			myErr = USBDeviceRequest(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": kSetRemoteWakeup - immediate error", myErr);
			}
			break;

		case kNewInterfaceRef:
			InitParamBlock(pCompositePB->deviceRef, &pCompositePB->pb);
			// Note: pCompositePB->usb.cntl.WIndex will be set to zero by InitParamBlock
			// so set it again to pCompositePB->interfaceIndex before calling USBNewInterfaceRef
			pCompositePB->pb.usb.cntl.WIndex = pCompositePB->interfaceIndex;
			pCompositePB->pb.usbRefcon |= kCompletionPending;
			pCompositePB->pb.usbCompletion = (USBCompletion)CompositeDeviceCompletionProc;

			myErr = USBNewInterfaceRef(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": kNewInterfaceRef - immediate error", myErr);
			}
			break;
			
		default:
			USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName" - Transaction initiated with bad refcon value", pCompositePB->pb.usbRefcon);
			pCompositePB->pb.usbRefcon = kUndefined + kReturnFromDriver;
			break;
	}
	
// At this point the control is returned to the system.  If a USB transaction
// has been initiated, then it will call the Complete procs
// (below) to handle the results of the transaction.
}

void CompositeDeviceCompletionProc(USBPB *pb)
{
OSStatus myErr;
register usbCompositePBStruct *pCompositePB;
USBInterfaceDescriptorPtr pInterfaceDescriptor;
UInt32 i;

	pCompositePB = (usbCompositePBStruct *)(pb);

	if (pCompositePB->expertTerminatePending)				// TCC <USB16>
		return;
		
	pCompositePB->transDepth--;
	if ((pCompositePB->transDepth < 0) || (pCompositePB->transDepth > 1))
	{
		USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName" - Illegal Transaction Depth", pCompositePB->transDepth);
	}
	
	if ((pCompositePB->pb.usbStatus == kUSBEndpointStallErr) && 
		(pCompositePB->pb.usbRefcon & ~(kCompletionPending + kReturnFromDriver)) == kSetRemoteWakeup){
		USBExpertStatus(pCompositePB->pb.usbReference, kCompositeDriverName": Device does not support RemoteWakeup (Config descriptor is wrong)", pCompositePB->partialConfigDescriptor.attributes);
		pCompositePB->pb.usbStatus = kUSBNoErr;
	}
	
	if((pCompositePB->pb.usbStatus != kUSBNoErr) && (pCompositePB->pb.usbStatus != kUSBPending))
	{
		USBExpertStatus(pCompositePB->pb.usbReference, kCompositeDriverName": Completion Error", pCompositePB->pb.usbStatus);
		pCompositePB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);
		pCompositePB->pb.usbRefcon |= kRetryTransaction;
		pCompositePB->retryCount--;
		if (!pCompositePB->retryCount)
		{
			USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": Too many retries", pCompositePB->pb.usbRefcon);
			pCompositePB->pb.usbRefcon = kReturnFromDriver;
			return;
		}
	}
	else
	{
		pCompositePB->pb.usbRefcon &= ~kRetryTransaction;
		pCompositePB->retryCount = kCompositeRetryCount;
	}

	if (pCompositePB->pb.usbRefcon & kCompletionPending)			 
	{												
		pCompositePB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);
		switch(pCompositePB->pb.usbRefcon)
		{
			case kResetDevice:			/* Device is not reset */
				pCompositePB->pb.usbRefcon = kGetFullConfiguration0;
				break;

			case kGetFullConfiguration0:
				pCompositePB->pFullConfigDescriptor = pCompositePB->pb.usbBuffer;
				if (pCompositePB->pFullConfigDescriptor == nil)
				{
					USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName": USBGetFullConfiguration - pointer is nil", pCompositePB->pb.usbRefcon);
					pCompositePB->pb.usbRefcon = kReturnFromDriver;
					break;
				}
				
				BlockCopy(   (void *)pCompositePB->pFullConfigDescriptor, (void *)(&(pCompositePB->partialConfigDescriptor)), (Size)(sizeof(USBConfigurationDescriptor) )  );
				USBExpertStatus(pCompositePB->pb.usbReference, kCompositeDriverName" - Configuration->maxPower = ", pCompositePB->pFullConfigDescriptor->maxPower);
				USBExpertStatus(pCompositePB->pb.usbReference, kCompositeDriverName" -      Hub supplied power = ", pCompositePB->busPowerAvailable);
				if (pCompositePB->pFullConfigDescriptor->maxPower <= pCompositePB->busPowerAvailable)
				{
					pCompositePB->pb.usbRefcon = kSetConfig;
				}
				else
				{
					USBExpertSetDevicePowerStatus(pCompositePB->pb.usbReference, 0, 0, kUSBDevicePower_BusPowerInsufficient, pCompositePB->busPowerAvailable, pCompositePB->pFullConfigDescriptor->maxPower);
					USBExpertFatalError(pCompositePB->pb.usbReference, kUSBDevicePowerProblem, kCompositeDriverName" - Insufficient power for device", pCompositePB->busPowerAvailable);
					pCompositePB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kSetConfig:			/* get all the interface descriptors and save them */
				for (i=0; i < pCompositePB->partialConfigDescriptor.numInterfaces; i++)
				{
					pCompositePB->interfaceRefArray[i] = 0;
					myErr = GetInterfaceDescriptor(pCompositePB->pFullConfigDescriptor, (UInt32)i, &pInterfaceDescriptor);
					if (kUSBNoErr == myErr)
					{
						BlockCopy((void *)pInterfaceDescriptor, (void *)(&(pCompositePB->interfaceDescriptors[i])), (Size)(pInterfaceDescriptor->length));
					}
					else				/* if GetInterfaceDescriptor returned an error, then set the length to zero (so we know later) */
					{
						pCompositePB->interfaceDescriptors[i].length = 0;
					}
				}
				pCompositePB->interfaceIndex = 0;
				pCompositePB->interfaceCount = pCompositePB->partialConfigDescriptor.numInterfaces;
				pCompositePB->pb.usbRefcon = 
					(pCompositePB->partialConfigDescriptor.attributes & kUSBAtrRemoteWakeup) ? kSetRemoteWakeup : kNewInterfaceRef;
				break;

			case kSetRemoteWakeup:			
				pCompositePB->pb.usbRefcon = kNewInterfaceRef;
				break;
				
			case kNewInterfaceRef:
				/* save the new interface ref for this interface */
				pCompositePB->interfaceRefArray[pCompositePB->interfaceIndex] = pCompositePB->pb.usbReference;
				
				/* only install the interface driver if we had a valid interface descriptor */
				if (pCompositePB->interfaceDescriptors[pCompositePB->interfaceIndex].length != 0)
				{
					USBExpertInstallInterfaceDriver(pCompositePB->interfaceRefArray[pCompositePB->interfaceIndex], &pCompositePB->deviceDescriptor, &pCompositePB->interfaceDescriptors[pCompositePB->interfaceIndex], pCompositePB->deviceRef, 0);
				}
				
				/* advance to the next interface */
				pCompositePB->interfaceIndex++;
				
				/* if there's more interfaces, then just keep cycling through kNewInterfaceRef */
				if (pCompositePB->interfaceIndex < pCompositePB->interfaceCount)
				{
					pCompositePB->pb.usbRefcon = kNewInterfaceRef;
				}
				else
				{
					pCompositePB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
			
			default:
				USBExpertFatalError(pCompositePB->pb.usbReference, kUSBInternalErr, kCompositeDriverName" - Transaction completed with a bad refcon value", pCompositePB->pb.usbRefcon);
				pCompositePB->pb.usbRefcon = kReturnFromDriver;
				break;
		}
	}
	if (!(pCompositePB->pb.usbRefcon & kReturnFromDriver) && (!pCompositePB->driverRemovalPending))
		CompositeDeviceInitiateTransaction(pb);

}

// entry points, moved from CompositeDriverDescription.c

// hubDriverInitInterface function
// Called to initialize driver for an individual interface - either by expert or
// internally by driver
OSStatus CompositeDriverInitInterface(
			UInt32 						interfaceNum, 
			USBInterfaceDescriptor		*interfaceDesc, 
			USBDeviceDescriptor			*deviceDesc, 
			USBDeviceRef 				device)
{
#pragma unused (interfaceNum)
#pragma unused (interfaceDesc)
#pragma unused (deviceDesc)
#pragma unused (device)

	return (OSStatus)kUSBNoErr;
}

OSStatus	CompositeDriverNotifyProc(UInt32 	notification, void *pointer, UInt32 refcon)
{
#pragma unused (pointer)
#pragma unused (refcon)

OSStatus	myErr;
OSStatus 	status = kUSBNoErr;
UInt32		count = 0;
static	UInt32	interfacenum = 0;

	switch (notification)
	{
		case kNotifyExpertTerminating:								// TCC <USB15>
				newInterfacesPB.expertTerminatePending = true;
				while (count < newInterfacesPB.interfaceCount)
						USBExpertRemoveInterfaceDriver(newInterfacesPB.interfaceRefArray[count++]);
				return(noErr);
			break;
		case kNotifyDriverBeingRemoved:
			newInterfacesPB.driverRemovalPending = true;
			
			if (newInterfacesPB.pb.usbRefcon & kCompletionPending)
			{
				USBExpertStatus(newInterfacesPB.deviceRef, kCompositeDriverName": Waiting for transaction to complete", newInterfacesPB.pb.usbRefcon);
				status = kUSBDeviceBusy;
			}
			else
			{
				if (interfacenum == 0)
					newInterfacesPB.pb.usbStatus = kUSBNoErr;
				
				if (newInterfacesPB.pb.usbStatus == kUSBPending)
				{
					USBExpertStatus(newInterfacesPB.deviceRef, kCompositeDriverName": Waiting for Dispose/Deallocate to complete", newInterfacesPB.pb.usbRefcon);
					status = kUSBDeviceBusy;
				}
				else
				{
					if (interfacenum < newInterfacesPB.interfaceCount)		
					{
						USBExpertRemoveInterfaceDriver(newInterfacesPB.interfaceRefArray[interfacenum]);
						InitParamBlock(newInterfacesPB.interfaceRefArray[interfacenum], &newInterfacesPB.pb);
						newInterfacesPB.pb.usbRefcon = 0; 			
						newInterfacesPB.pb.usbCompletion = (USBCompletion)kUSBNoCallBack;
						myErr = USBDisposeInterfaceRef(&newInterfacesPB.pb);
						if(immediateError(myErr))
						{
							USBExpertFatalError(newInterfacesPB.interfaceRefArray[interfacenum], kUSBInternalErr, kCompositeDriverName" - USBDisposeInterfaceRef - immediate error", myErr);
						}
						interfacenum++;
						status = kUSBDeviceBusy;
					}
					else
					{
						if (newInterfacesPB.pFullConfigDescriptor)
						{
							InitParamBlock(newInterfacesPB.deviceRef, &newInterfacesPB.pb);
							newInterfacesPB.pb.usbRefcon = 0; 			
							newInterfacesPB.pb.usbBuffer = newInterfacesPB.pFullConfigDescriptor;		
							newInterfacesPB.pb.usbCompletion = (USBCompletion)kUSBNoCallBack;
							myErr = USBDeallocMem(&newInterfacesPB.pb);
							newInterfacesPB.pFullConfigDescriptor = nil;
							status = kUSBDeviceBusy;
						}
						else
						{
							status = kUSBNoErr;
						}
					}
				}
			}
			break;
			
		default:
			break;
	}
	return status;
}

// Hardware Validation
// Called upon load by Expert
OSStatus CompositeDriverValidateHW(USBDeviceRef device, USBDeviceDescriptor *desc)
{
#pragma unused (device)
#pragma unused (desc)

	return (OSStatus)kUSBNoErr;
}

// Initialization function
// Called upon load by Expert
OSStatus 	CompositeDriverInitialize (USBDeviceRef device, USBDeviceDescriptorPtr pDesc,  UInt32 busPowerAvailable)
{
	// don't let a race condition occur; prevent finalizing until we're all done.
	DeviceInitialize(device, pDesc, busPowerAvailable);
	return (OSStatus)kUSBNoErr;
}

// Termination function
// Called by Expert when driver is being shut down
OSStatus CompositeDriverFinalize(USBDeviceRef theDeviceRef, USBDeviceDescriptorPtr pDesc)
{
#pragma unused (pDesc)

	USBExpertStatus(theDeviceRef, kCompositeDriverName" - Finalize", 0);
	return (OSStatus)kUSBNoErr;
}

void DeviceInitialize(USBDeviceRef device, USBDeviceDescriptorPtr pDeviceDescriptor, UInt32 busPowerAvailable)
{
static Boolean beenThereDoneThat = false;

	if(beenThereDoneThat)
	{
		USBExpertFatalError(device, kUSBInternalErr, kCompositeDriverName" is not reentrant!", 0);
		return;
	}
	beenThereDoneThat = true;
	
	newInterfacesPB.driverRemovalPending = false;
	newInterfacesPB.expertTerminatePending = false;
	
	newInterfacesPB.deviceRef = device;	
	newInterfacesPB.deviceDescriptor = *pDeviceDescriptor;	
	
	newInterfacesPB.busPowerAvailable = busPowerAvailable;							
	newInterfacesPB.delayLevel = 0;							
	newInterfacesPB.transDepth = 0;							
	newInterfacesPB.retryCount = kCompositeRetryCount;
	newInterfacesPB.pFullConfigDescriptor = nil;
	
	InitParamBlock(device, &newInterfacesPB.pb);
	newInterfacesPB.pb.usbRefcon = kResetDevice;			/* Start out at first state */
	
//	DebugStr("\pIn Composite Driver");
	CompositeDeviceInitiateTransaction(&newInterfacesPB.pb);
}


