/*
	File:		USBKeypad.c

	Contains:	xxx put contents here xxx

	Version:	xxx put version here xxx

	Copyright:	 1999 by Apple Computer, Inc., all rights reserved.

*/

#include <Types.h>
#include <Devices.h>
#include <processes.h>
#include <DriverServices.h>
#include <USB.h>

#include "USBKeypad.h"

usbKeyPadPBStruct myParamBlock;
usbKeyPadPBStruct shimParamBlock;


void InitParamBlock(USBDeviceRef theDeviceRef, USBPB * paramblock);

void InitParamBlock(USBDeviceRef theDeviceRef, USBPB * paramblock)
{
	paramblock->usbReference = theDeviceRef;
	paramblock->pbVersion = kUSBCurrentPBVersion;
	paramblock->usb.cntl.WIndex = 0; 			
	paramblock->usbBuffer = nil;		
	paramblock->usbStatus = noErr;
	paramblock->usbReqCount = 0;
	paramblock->usb.cntl.WValue = 0;
	paramblock->usbFlags = 0;
	paramblock->usbOther = 0;
}

static Boolean immediateError(OSStatus err)
{
	return((err != kUSBPending) && (err != noErr) );
}

void InitiateTransaction(USBPB *pb)
{
register usbKeyPadPBStruct *pKeyPadPB;
OSStatus myErr;

	pKeyPadPB = (usbKeyPadPBStruct *)(pb);
	pKeyPadPB->transDepth++;
	if (pKeyPadPB->transDepth < 0)
	{
		USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: transDepth < 0 (initiation)", pKeyPadPB->pb.usbRefcon );
	}
	
	if (pKeyPadPB->transDepth > 1)
	{
		USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: transDepth > 1 (initiation)", pKeyPadPB->pb.usbRefcon );
	}
	
	switch(pKeyPadPB->pb.usbRefcon & ~kRetryTransaction)
	{
		case kFindInterface:
			InitParamBlock(pKeyPadPB->deviceRef, &pKeyPadPB->pb);
			pKeyPadPB->pb.usb.cntl.WIndex = 0;
			pKeyPadPB->pb.usb.cntl.WValue = 0;
			pKeyPadPB->pb.usbClassType = kUSBHIDInterfaceClass;
			pKeyPadPB->pb.usbSubclass = 0;
			pKeyPadPB->pb.usbProtocol = kUSBKeyboardInterfaceProtocol;
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBFindNextInterface(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->pb.usbReference, kUSBInternalErr, "\pTenKey Driver: USBFindNextInterface (immediate error)", myErr);
				pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kConfigureDevice:
			InitParamBlock(pKeyPadPB->deviceRef, &pKeyPadPB->pb);
			pKeyPadPB->pb.usb.cntl.WValue = pKeyPadPB->configurationNumber;
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBOpenDevice(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->pb.usbReference, kUSBInternalErr, "\pTenKey Driver: USBOpenDevice (immediate error)", myErr);
				pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kNewInterfaceRef:
			InitParamBlock(pKeyPadPB->deviceRef, &pKeyPadPB->pb);
			pKeyPadPB->pb.usb.cntl.WIndex = pKeyPadPB->interfaceNumber;
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBNewInterfaceRef(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->pb.usbReference, kUSBInternalErr, "\pTenKey Driver: USBNewInterfaceRef (immediate error)", myErr);
				pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kConfigureInterface:
			InitParamBlock(pKeyPadPB->interfaceRef, &pKeyPadPB->pb);
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBConfigureInterface(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->pb.usbReference, kUSBInternalErr, "\pTenKey Driver: USBConfigureInterface (immediate error)", myErr);
				pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kGetReportDescriptor:
			InitParamBlock(pKeyPadPB->deviceRef, &pKeyPadPB->pb);
			pKeyPadPB->pb.usb.cntl.BMRequestType = USBMakeBMRequestType(kUSBIn, kUSBStandard, kUSBInterface);
			pKeyPadPB->pb.usb.cntl.BRequest = kUSBRqGetDescriptor;
			pKeyPadPB->pb.usb.cntl.WValue = (0x22 << 8) + 1; 			// get the first report descriptor
			pKeyPadPB->pb.usb.cntl.WIndex = pKeyPadPB->interfaceNumber;
			pKeyPadPB->pb.usbReqCount = 0x66;
			pKeyPadPB->pb.usbBuffer = pKeyPadPB->hidReportDescriptor;
			
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBDeviceRequest(pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->pb.usbReference, kUSBInternalErr, "\pTenKey Driver: kGetReportDescriptor - immediate error", myErr);
			}
			break;
		
		case kFindInterruptPipe:
			USBExpertStatus(pKeyPadPB->interfaceRef, "\pTenKey Driver: kFindInterruptPipe - interfaceRef", pKeyPadPB->interfaceRef);
			InitParamBlock(pKeyPadPB->interfaceRef, &pKeyPadPB->pb);
			pKeyPadPB->pb.usbFlags = kUSBIn;
			pKeyPadPB->pb.usbClassType = kUSBInterrupt; 
			
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBFindNextPipe(&pKeyPadPB->pb);
			if (immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: kFindInterruptPipe (immediate error)", myErr);
				pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
			}
			break;
			
		case kReadInterruptPipe:
			InitParamBlock(pKeyPadPB->pipeRef, &pKeyPadPB->pb);

			pKeyPadPB->pb.usbBuffer = (Ptr)pKeyPadPB->hidReport;
			pKeyPadPB->pb.usbReqCount = 4;
			pKeyPadPB->pb.usb.cntl.WIndex = pKeyPadPB->interfaceNumber;	
			
			pKeyPadPB->pb.usbRefcon |= kCompletionPending;
			pKeyPadPB->pb.usbCompletion = (USBCompletion)TransactionCompletionProc;
			myErr = USBIntRead(&pKeyPadPB->pb);
			if(immediateError(myErr))
			{
				USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: kReadInterruptPipe (ImmediateError)", myErr);
			}
			break;
			
			
		default:
			USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: Transaction completed with bad refcon value", pKeyPadPB->pb.usbRefcon );
			pKeyPadPB->pb.usbRefcon = kUndefined + kReturnFromDriver;
			break;
	}
	
// At this point the control is returned to the system.  If a USB transaction
// has been initiated, then it will call the Complete procs
// (below) to handle the results of the transaction.
}

void TransactionCompletionProc(USBPB *pb)
{
unsigned char	* errstring;
register usbKeyPadPBStruct *pKeyPadPB;
USBPipeState pipeState;

	pKeyPadPB = (usbKeyPadPBStruct *)(pb);
	pKeyPadPB->transDepth--;
	if (pKeyPadPB->transDepth < 0)
	{
		USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: transDepth < 0 (completion)", pKeyPadPB->pb.usbRefcon );
	}
	
	if (pKeyPadPB->transDepth > 1)
	{
		USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: transDepth > 1 (completion)", pKeyPadPB->pb.usbRefcon );
	}
	
	if(pKeyPadPB->pb.usbStatus != noErr)													// was there an error?
	{
		switch(pKeyPadPB->pb.usbRefcon & 0x0fff)											// yes, so show where the error occurred
		{
			case kFindInterface:  				errstring = "\pTenKey Driver: Error during kFindInterface"; break;
			case kConfigureDevice:				errstring = "\pTenKey Driver: Error during kConfigureDevice"; break;
			case kNewInterfaceRef:				errstring = "\pTenKey Driver: Error during kNewInterfaceRef"; break;
			case kConfigureInterface:  			errstring = "\pTenKey Driver: Error during kConfigureInterface"; break;
			case kGetReportDescriptor: 	 		errstring = "\pTenKey Driver: Error during kGetReportDescriptor"; break;
			case kFindInterruptPipe: 	 		errstring = "\pTenKey Driver: Error during kFindInterruptPipe"; break;
			case kReadInterruptPipe: 	 		errstring = "\pTenKey Driver: Error during kReadInterruptPipe"; break;
		};
		USBExpertFatalError(pKeyPadPB->deviceRef, pKeyPadPB->pb.usbStatus, errstring, (pKeyPadPB->pb.usbRefcon & 0x0fff));
		
		pKeyPadPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);				// set up to retry the transaction
		pKeyPadPB->pb.usbRefcon |= kRetryTransaction;
		pKeyPadPB->retryCount--;
		
		if ((!pKeyPadPB->retryCount)	|| (pKeyPadPB->pb.usbStatus == kUSBAbortedError))	// have we exhausted the retries?
		{																					// or received an abort?
			USBExpertStatus(pKeyPadPB->deviceRef, "\pTenKey Driver: Pipe abort or unable to recover from error", pKeyPadPB->deviceRef);
			pKeyPadPB->pb.usbRefcon = kReturnFromDriver;									// if so, just exit.
		}
		else																				// if it didn't abort and there's retries left, then...
		{
			if (pKeyPadPB->pipeRef)														// check if the pipe is open.
			{
				USBGetPipeStatusByReference(pKeyPadPB->pipeRef, &pipeState);				// yes, so what it's state?
				if (pipeState != kUSBActive)												// if it's not active, try to clear it.  It might be stalled...
				{
					USBExpertStatus(pKeyPadPB->deviceRef, "\pTenKey Driver: Pipe is open and stalled, clearing stall...", pKeyPadPB->deviceRef);
					USBClearPipeStallByReference(pKeyPadPB->pipeRef);
				}
			}
		}
	}
	else
	{
		pKeyPadPB->pb.usbRefcon &= ~kRetryTransaction;
		pKeyPadPB->retryCount = kKeyboardRetryCount;
	}

	if (pKeyPadPB->pb.usbRefcon & kCompletionPending)			 
	{												
		pKeyPadPB->pb.usbRefcon &= ~(kCompletionPending + kReturnFromDriver);
		switch(pKeyPadPB->pb.usbRefcon)
		{
			case kFindInterface:
				if ((pKeyPadPB->pb.usbClassType == kUSBHIDInterfaceClass) && 
					(pKeyPadPB->pb.usbProtocol == kUSBKeyboardInterfaceProtocol))
				{
					pKeyPadPB->pb.usbRefcon = kConfigureDevice;
					pKeyPadPB->interfaceNumber = pKeyPadPB->pb.usb.cntl.WIndex;
					pKeyPadPB->configurationNumber = pKeyPadPB->pb.usb.cntl.WValue;
					pKeyPadPB->interfaceSubClass = pKeyPadPB->pb.usbSubclass;
				}
				else
				{
					USBExpertStatus(pKeyPadPB->deviceRef, "\pTenKey Driver: Keyboard interface was not found", pKeyPadPB->deviceRef);
					pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kConfigureDevice:
				pKeyPadPB->pb.usbRefcon = kNewInterfaceRef;
				break;
				
			case kNewInterfaceRef:
				pKeyPadPB->interfaceRef = pKeyPadPB->pb.usbReference;
				pKeyPadPB->pb.usbRefcon = kConfigureInterface;
				break;
			
			case kConfigureInterface:
				pKeyPadPB->pb.usbRefcon = kFindInterruptPipe;
				break;
			
			case kGetReportDescriptor:
				pKeyPadPB->pb.usbRefcon = kFindInterruptPipe;
				break;
				
			case kFindInterruptPipe:
				if ((pKeyPadPB->pb.usbClassType == kUSBInterrupt) && 
					(pKeyPadPB->pb.usbFlags == kUSBIn))
				{
					pKeyPadPB->hidReportSize = pKeyPadPB->pb.usb.cntl.WValue;
					pKeyPadPB->pipeRef = pKeyPadPB->pb.usbReference;
					pKeyPadPB->pb.usbRefcon = kReadInterruptPipe;
				}
				else
				{
					pKeyPadPB->pb.usbRefcon = kReturnFromDriver;
				}
				break;
				
			case kReadInterruptPipe:
				//DebugStr("\pKeypad read completed");
				ProcessInterruptReport(pKeyPadPB->hidReport);
				pKeyPadPB->pb.usbRefcon = kReadInterruptPipe;
				break;
				
			default:
				USBExpertFatalError(pKeyPadPB->deviceRef, kUSBInternalErr, "\pTenKey Driver: Transaction completed with bad refcon value", pKeyPadPB->pb.usbRefcon );
				pKeyPadPB->pb.usbRefcon = kUndefined + kReturnFromDriver;
				break;
		}
	}
	if (!(pKeyPadPB->pb.usbRefcon & kReturnFromDriver))
		InitiateTransaction(pb);
}


void DriverEntry(USBDeviceRef device, USBDeviceDescriptorPtr pDesc, UInt32 busPowerAvailable)
{
#pragma unused (busPowerAvailable)

static Boolean beenThereDoneThat = false;

	if(beenThereDoneThat)
	{
		USBExpertFatalError(device, kUSBInternalErr, "\pTenKey is not reentrant", 12);
		return;
	}
	beenThereDoneThat = true;
	
//	DebugStr("\pIn Key Pad Interface entry routine");
	shimParamBlock.deviceDescriptor = *pDesc;				
	shimParamBlock.transDepth = 0;							
	shimParamBlock.retryCount = kKeyboardRetryCount;
	shimParamBlock.deviceRef = device;
	shimParamBlock.interfaceRef = 0;
	shimParamBlock.pipeRef = 0;
	
	shimParamBlock.deviceRef = device;		
	shimParamBlock.interfaceRef = nil;		
	shimParamBlock.pipeRef = nil;		
	
	InitParamBlock(device, &shimParamBlock.pb);
	shimParamBlock.pb.pbLength = sizeof(usbKeyPadPBStruct);
	shimParamBlock.pb.usbRefcon = 0;				

	myParamBlock.deviceDescriptor = *pDesc;				
	myParamBlock.transDepth = 0;							
	myParamBlock.retryCount = kKeyboardRetryCount;
	myParamBlock.deviceRef = device;
	myParamBlock.interfaceRef = 0;
	myParamBlock.pipeRef = 0;

	myParamBlock.deviceRef = device;		
	myParamBlock.interfaceRef = nil;		
	myParamBlock.pipeRef = nil;	
	
	myParamBlock.sendRawReportFlag = false;
	myParamBlock.hidEmulationInit = false;
	
	InitParamBlock(device, &myParamBlock.pb);
	myParamBlock.pb.pbLength = sizeof(usbKeyPadPBStruct);
	myParamBlock.pb.usbRefcon = kFindInterface;		/* Start with setting the interface protocol */
	
	
	USBHIDControlDevice(kHIDEnableDemoMode,0);
	InitiateTransaction(&myParamBlock.pb);
}



